from torch import nn
import torch.nn.functional as F
import torch
import math
class Net(nn.Module):

    def __init__(self, hdim = 64, ch=48):
        super(Net,self).__init__()

        self.hdim = hdim
        self.ch = ch
        
        self.conv1 = nn.Conv2d(1,ch,kernel_size=5,stride=1).cuda()
        self.conv2 = nn.Conv2d(ch,ch,kernel_size=5,stride=1).cuda()
        self.pool1 = nn.MaxPool2d(kernel_size=2,stride=2) #2x2 maxpool
        self.pool2 = nn.MaxPool2d(kernel_size=2,stride=2) #3x3 maxpool
        self.fc1 = nn.Linear(4*4*ch,hdim).cuda()
        self.fc2 = nn.Linear(hdim,10).cuda()
                
        self.conv1bn=self.makebn(2,ch,5*5*ch)
        self.conv2bn=self.makebn(2,ch,4*4*ch)
        self.fc1bn=self.makebn(1,hdim,hdim)
        self.fc2bn=self.makebn(1,10,10)
        
        print('Parameters:',sum(p.numel() for p in self.parameters()))
        print('Learnable Parameters:',sum(p.numel()*p.requires_grad for p in self.parameters()))

    def makebn(self, n, a, b):
        #define batchnorm_nD with channel #  and normalize by 1/b
        if n==1:
            BN=nn.BatchNorm1d(a).cuda()
        if n==2:
            BN=nn.BatchNorm2d(a).cuda()
        for weight in BN.parameters():
            with torch.no_grad():
                weight.div_(b**.5)            
            weight.requires_grad=False        
        return BN

    def normalize(self):
        with torch.no_grad():
            R = self.paranorm()
            for p in self.parameters():
                if p.requires_grad == True:
                    p1 = p/R
                    p.copy_(p1)

    def gradnorm(self):
        S=0.
        with torch.no_grad():
            for module in (self.conv1,self.conv2,self.fc1,self.fc2):
                S+=sum((p.grad*p.grad).sum() for p in module.parameters())
        return S**.5

    def paranorm(self):
        S=0.
        with torch.no_grad():
            for module in (self.conv1,self.conv2,self.fc1,self.fc2):
                S+=sum((p*p).sum() for p in module.parameters())
        return S**.5


    def compare(self,other):
        S=0.
        with torch.no_grad():
            for p1, p2 in zip(self.parameters(), other.parameters()):
                S+=((p1-p2)*(p1-p2)*p1.requires_grad).sum()
        return S**.5
                    
                
    def flow(self,lr = 0):
        with torch.no_grad():
            for p in self.parameters():
                if p.requires_grad == True:
                    p1 = p-lr*(p.grad)
                    p.copy_(p1)
                
    def seq(self,x):
        #print(self.conv1(x).size())
        x = F.relu(self.conv1bn(self.conv1(x))) #24x24x10
        x = self.pool1(x) #12x12x10
        x = F.relu(self.conv2bn(self.conv2(x))) #8x8x10
        x = self.pool2(x) #4x4x10    
        x = x.view(-1, 4*4*self.ch) #flattening
        x = F.relu(self.fc1bn(self.fc1(x)))
        x = self.fc2bn(self.fc2(x))
        return x

    def state(self):
        l = (self.conv1.weight,self.conv1.bias,self.conv2.weight,self.conv2.bias,\
             self.fc1.weight,self.fc1.bias,self.fc2.weight,self.fc2.bias)
        x = torch.empty(0).cuda()
        for p in l:
            x=torch.cat((x,p.view(-1)))
        return x
  
    def forward(self,X,target,Criterion):
        bs=X.size()[0]

        h=self.seq(X)
        target=target.long()

        L_total=Criterion(h,target)
                    #nn.CrossEntropyLoss is the sum of CrossEntropyLoss for y(index=1..batchsize),
                    #and pred(1..batchsize) divided by batchsize (500 here)
        A_total=(torch.max(h,1).indices==target).float()
        
        L=L_total.sum()
        A=A_total.sum()/bs
        return L,A,
            
def run(net,X,y,Criterion,lr, train=True):
    
    G=torch.zeros(1)
    
    if train==True:
        L,A=net(X,y,Criterion)
        L.backward(retain_graph=True)
        G = net.gradnorm()
        net.flow(lr)
        net.normalize()
        net.zero_grad()
        
    if train==False:
        with torch.no_grad():
            L,A = net(X,y,Criterion)

    return L.detach(), A.detach(), G.detach()

def Run(net,DL,Criterion,lr, train=True):

    Loss, Accu, Grad= 0, 0, 0
    for batchcounter,(X,y) in enumerate(DL):
        L, A, G=run(net,X,y,Criterion,lr,train)
        Loss+=L.item()
        Accu+=A.item()
        Grad+=G.item()
    Loss /= len(DL)
    Accu /= len(DL)
    Grad /= len(DL)
    return Loss, Accu, Grad

